#! /usr/bin/env python
#
# Copyright 2021-2026 Spotify AB
#
# Licensed under the GNU Public License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.gnu.org/licenses/gpl-3.0.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = "0.9.23"
# Done this way so we can use dynamic-metadata to parse the version above.
MAJOR, MINOR, PATCH = (int(x) for x in __version__.split("."))



#! /usr/bin/env python
#
# Copyright 2021 Spotify AB
#
# Licensed under the GNU Public License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.gnu.org/licenses/gpl-3.0.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# In certain situations (including when using asyncio), Pedalboard's C++ code will
# implicitly and lazily import NumPy via its Pybind11 bindings. This lazy import
# can sometimes (!) cause a deadlock on the GIL if multiple threads are attempting
# it simultaneously. To work around this, we preload Numpy into the current Python
# process so that Pybind11's lazy import is effectively a no-op.
import numpy  # noqa

try:
    from pedalboard_native import *  # noqa: F403, F401
except ImportError as _e:
    if "DLL load failed" in str(_e):
        # On Windows, we get an obtuse "DLL load failed" if the
        # MSVC++ redistributable library isn't installed
        raise ImportError(
            "Pedalboard failed to load its native code. Hint: try installing the Microsoft Visual"
            " C++ Redistributable."
        )
    else:
        raise
from pedalboard_native.utils import *  # noqa: F403, F401 # type: ignore

from ._pedalboard import (  # noqa: F401
    _AVAILABLE_PLUGIN_CLASSES,
    AudioProcessorParameter,  # noqa: F401
    ExternalPlugin,  # noqa: F401
    Pedalboard,  # noqa: F401
    load_plugin,  # noqa: F401
)

# noqa: F401
from .version import __version__  # noqa: F401

for _klass in _AVAILABLE_PLUGIN_CLASSES:
    vars()[_klass.__name__] = _klass



from __future__ import annotations
import pedalboard_native.utils

import typing

original_overload = typing.overload
__OVERLOADED_DOCSTRINGS = {}

def patch_overload(func):
    original_overload(func)
    if func.__doc__:
        __OVERLOADED_DOCSTRINGS[func.__qualname__] = func.__doc__
    else:
        func.__doc__ = __OVERLOADED_DOCSTRINGS.get(func.__qualname__)
    if func.__doc__:
        # Work around the fact that pybind11-stubgen generates
        # duplicate docstrings sometimes, once for each overload:
        docstring = func.__doc__
        if docstring[len(docstring) // 2 :].strip() == docstring[: -len(docstring) // 2].strip():
            func.__doc__ = docstring[len(docstring) // 2 :].strip()
    return func

typing.overload = patch_overload

from typing_extensions import Literal
from enum import Enum
import threading
import numpy
import pedalboard_native

_Shape = typing.Tuple[int, ...]

__all__ = ["Chain", "Mix", "time_stretch"]

class Chain(pedalboard_native.PluginContainer, pedalboard_native.Plugin):
    """
    Run zero or more plugins as a plugin. Useful when used with the Mix plugin.
    """

    @typing.overload
    def __init__(self, plugins: typing.List[pedalboard_native.Plugin]) -> None: ...
    @typing.overload
    def __repr__(self) -> str: ...
    pass

class Mix(pedalboard_native.PluginContainer, pedalboard_native.Plugin):
    """
    A utility plugin that allows running other plugins in parallel. All plugins provided will be mixed equally.
    """

    @typing.overload
    def __init__(self, plugins: typing.List[pedalboard_native.Plugin]) -> None: ...
    @typing.overload
    def __repr__(self) -> str: ...
    pass

def time_stretch(
    input_audio: numpy.ndarray[typing.Any, numpy.dtype[numpy.float32]],
    samplerate: float,
    stretch_factor: typing.Union[
        float, numpy.ndarray[typing.Any, numpy.dtype[numpy.float64]]
    ] = 1.0,
    pitch_shift_in_semitones: typing.Union[
        float, numpy.ndarray[typing.Any, numpy.dtype[numpy.float64]]
    ] = 0.0,
    high_quality: bool = True,
    transient_mode: str = "crisp",
    transient_detector: str = "compound",
    retain_phase_continuity: bool = True,
    use_long_fft_window: typing.Optional[bool] = None,
    use_time_domain_smoothing: bool = False,
    preserve_formants: bool = True,
) -> numpy.ndarray[typing.Any, numpy.dtype[numpy.float32]]:
    """
    Time-stretch (and optionally pitch-shift) a buffer of audio, changing its length.

    Using a higher ``stretch_factor`` will shorten the audio - i.e., a ``stretch_factor``
    of ``2.0`` will double the *speed* of the audio and halve the *length* of the audio,
    without changing the pitch of the audio.

    This function allows for changing the pitch of the audio during the time stretching
    operation. The ``stretch_factor`` and ``pitch_shift_in_semitones`` arguments are
    independent and do not affect each other (i.e.: you can change one, the other, or both
    without worrying about how they interact).

    Both ``stretch_factor`` and ``pitch_shift_in_semitones`` can be either floating-point
    numbers or NumPy arrays of double-precision floating point numbers. Providing a NumPy
    array allows the stretch factor and/or pitch shift to vary over the length of the
    output audio.

    .. note::
        If a NumPy array is provided for ``stretch_factor`` or ``pitch_shift_in_semitones``:
          - The length of each array must be the same as the length of the input audio.
          - More frequent changes in the stretch factor or pitch shift will result in
            slower processing, as the audio will be processed in smaller chunks.
          - Changes to the ``stretch_factor`` or ``pitch_shift_in_semitones`` more frequent
            than once every 1,024 samples (23 milliseconds at 44.1kHz) will not have any
            effect.

    The additional arguments provided to this function allow for more fine-grained control
    over the behavior of the time stretcher:

      - ``high_quality`` (the default) enables a higher quality time stretching mode.
        Set this option to ``False`` to use less CPU power.

      - ``transient_mode`` controls the behavior of the stretcher around transients
        (percussive parts of the audio). Valid options are ``"crisp"`` (the default),
        ``"mixed"``, or ``"smooth"``.

      - ``transient_detector`` controls which method is used to detect transients in the
        audio signal. Valid options are ``"compound"`` (the default), ``"percussive"``,
        or ``"soft"``.

      - ``retain_phase_continuity`` ensures that the phases of adjacent frequency bins in
        the audio stream are kept as similar as possible. Set this to ``False`` for a
        softer, phasier sound.

      - ``use_long_fft_window`` controls the size of the fast-Fourier transform window
        used during stretching. The default (``None``) will result in a window size that
        varies based on other parameters and should produce better results in most
        situations. Set this option to ``True`` to result in a smoother sound (at the
        expense of clarity and timing), or ``False`` to result in a crisper sound.

      - ``use_time_domain_smoothing`` can be enabled to produce a softer sound with
        audible artifacts around sharp transients. This option mixes well with
        ``use_long_fft_window=False``.

      - ``preserve_formants`` allows shifting the pitch of notes without substantially
        affecting the pitch profile (formants) of a voice or instrument.

    .. warning::
        This is a function, not a :py:class:`Plugin` instance, and cannot be
        used in :py:class:`Pedalboard` objects, as it changes the duration of
        the audio stream.


    .. note::
        The ability to pass a NumPy array for ``stretch_factor`` and
        ``pitch_shift_in_semitones`` was added in Pedalboard v0.9.8.
    """



"""This module provides classes and functions for reading and writing audio files or streams.

*Introduced in v0.5.1.*"""

from __future__ import annotations
import pedalboard_native.io

import typing

from typing_extensions import Literal
import numpy as np
import pedalboard_native.utils
from numpy.typing import NDArray
from numpy import float32

_Shape = typing.Tuple[int, ...]

__all__ = [
    "AudioFile",
    "AudioStream",
    "ReadableAudioFile",
    "ResampledReadableAudioFile",
    "StreamResampler",
    "WriteableAudioFile",
    "WriteableAudioFileFlag",
    "get_supported_read_formats",
    "get_supported_write_formats",
]

class WriteableAudioFileFlag:
    """
    An enumeration of codec-specific options that can be passed when opening an
    audio file for writing. These flags are used as keys in the ``codec_options``
    dictionary parameter accepted by :class:`WriteableAudioFile` and
    :class:`AudioFile`.

    Not all flags are supported by all codecs. Passing an unsupported flag for
    the selected codec will raise a ``ValueError``.

    .. note::
        These flags control low-level encoder behavior. Most users will not need
        to use them. The ``quality`` parameter is usually sufficient for
        controlling encoder output.

    Members:

      Mp3EnableBitReservoir :
        When writing MP3 files, controls whether the LAME encoder's bit reservoir
        is enabled. Defaults to ``True``.
    """

    Mp3EnableBitReservoir: typing.ClassVar[WriteableAudioFileFlag]
    """
    When writing MP3 files, controls whether the LAME encoder's bit reservoir
    is enabled. The bit reservoir allows the encoder to use fewer bits on
    simple frames and save them for complex frames, improving overall quality
    at a given bitrate. Disabling it forces each frame to be independently
    decodable at the cost of slightly lower quality.

    Set to ``False`` for streaming or seeking applications where individual
    frames need to be independently decodable.

    Accepts a ``bool`` value. Defaults to ``True`` (bit reservoir enabled).
    """

    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class AudioFile:
    """
    A base class for readable and writeable audio files.

    :class:`AudioFile` may be used just like a regular Python ``open``
    function call, to open an audio file for reading (with the default ``"r"`` mode)
    or for writing (with the ``"w"`` mode).

    Unlike a typical ``open`` call:
     - :class:`AudioFile` objects can only be created in read (``"r"``) or write (``"w"``) mode.
       All audio files are binary (so a trailing ``b`` would be redundant) and appending to an
       existing audio file is not possible.
     - If opening an audio file in write mode (``"w"``), one additional argument is required:
       the sample rate of the file.
     - A file-like object can be provided to :class:`AudioFile`, allowing for reading and
       writing to in-memory streams or buffers. The provided file-like object must be seekable
       and must be opened in binary mode (i.e.: ``io.BytesIO`` instead of ``io.StringIO``).
       A :class:`memoryview` object may also be provided when reading audio.


    Examples
    --------

    Opening an audio file on disk::

       with AudioFile("my_file.mp3") as f:
           first_ten_seconds = f.read(int(f.samplerate * 10))


    Opening a file-like object::

       ogg_buffer: io.BytesIO = get_audio_buffer(...)
       with AudioFile(ogg_buffer) as f:
           first_ten_seconds = f.read(int(f.samplerate * 10))


    Opening an audio file on disk, while resampling on-the-fly::

        with AudioFile("my_file.mp3").resampled_to(22_050) as f:
           first_ten_seconds = f.read(int(f.samplerate * 10))


    Writing an audio file on disk::

       with AudioFile("white_noise.wav", "w", samplerate=44100, num_channels=2) as f:
           f.write(np.random.rand(2, 44100))


    Writing encoded audio to a file-like object::

       wav_buffer = io.BytesIO()
       with AudioFile(wav_buffer, "w", samplerate=44100, num_channels=2, format="wav") as f:
           f.write(np.random.rand(2, 44100))
       wav_buffer.getvalue()  # do something with the file-like object


    Encoding audio as ``wav``, ``ogg``, ``mp3``, or ``flac`` as a :class:`bytes` buffer in one line::

       sr = 44100
       num_channels = 2
       audio = np.random.rand(num_channels, sr)

       wav_buffer = AudioFile.encode(audio, sr, num_channels, format="wav")
       ogg_buffer = AudioFile.encode(audio, sr, num_channels, format="ogg")
       mp3_buffer = AudioFile.encode(audio, sr, num_channels, format="mp3")
       flac_buffer = AudioFile.encode(audio, sr, num_channels, format="flac")


    Writing to an audio file while also specifying quality options for the codec::

       with AudioFile(
           "white_noise.mp3",
           "w",
           samplerate=44100,
           num_channels=2,
           quality=160,  # kilobits per second
       ) as f:
           f.write(np.random.rand(2, 44100))


    Re-encoding a WAV file as an MP3 in four lines of Python::

       with AudioFile("input.wav") as i:
           with AudioFile("output.mp3", "w", i.samplerate, i.num_channels) as o:
               while i.tell() < i.frames:
                   o.write(i.read(1024))


    .. note::
        Calling the :class:`AudioFile` constructor does not actually return an
        :class:`AudioFile`. If opening an audio file in read ("r") mode, a
        :class:`ReadableAudioFile` will be returned. If opening an audio file
        in write ("w") mode, a :class:`WriteableAudioFile` will be returned. See
        those classes below for documentation.
    """

    @classmethod
    @typing.overload
    def __new__(cls, filename: str) -> ReadableAudioFile:
        """Open an audio file for reading (mode 'r' is implied)."""
        ...

    @classmethod
    @typing.overload
    def __new__(cls, filename: str, mode: Literal["r"]) -> ReadableAudioFile:
        """Open an audio file for reading with an explicit mode 'r'."""
        ...

    @classmethod
    @typing.overload
    def __new__(
        cls, file_like: typing.Union[typing.BinaryIO, memoryview], mode: Literal["r"] = "r"
    ) -> ReadableAudioFile: ...

    @classmethod
    @typing.overload
    def __new__(
        cls,
        filename: str,
        mode: Literal["w"],
        samplerate: typing.Optional[float] = None,
        num_channels: int = 1,
        bit_depth: int = 16,
        quality: typing.Optional[typing.Union[str, float]] = None,
        codec_options: typing.Dict[WriteableAudioFileFlag, typing.Union[bool, float, int, str]] = ...,
    ) -> WriteableAudioFile: ...

    @classmethod
    @typing.overload
    def __new__(
        cls,
        file_like: typing.BinaryIO,
        mode: Literal["w"],
        samplerate: typing.Optional[float] = None,
        num_channels: int = 1,
        bit_depth: int = 16,
        quality: typing.Optional[typing.Union[str, float]] = None,
        format: typing.Optional[str] = None,
        codec_options: typing.Dict[WriteableAudioFileFlag, typing.Union[bool, float, int, str]] = ...,
    ) -> WriteableAudioFile: ...
    
    @staticmethod
    def encode(
        samples: NDArray[typing.Union[np.int8, np.int16, np.int32, np.float32, np.float64]],
        samplerate: float,
        format: str,
        num_channels: int = 1,
        bit_depth: int = 16,
        quality: typing.Optional[typing.Union[str, float]] = None,
        codec_options: typing.Dict[WriteableAudioFileFlag, typing.Union[bool, float, int, str]] = ...,
    ) -> bytes:
        """
        Encode an audio buffer to a Python :class:`bytes` object.

        This function will encode an entire audio buffer at once and return a :class:`bytes`
        object representing the bytes of the resulting audio file.

        This function produces identical output to the following code::

            buf = io.BytesIO()
            with AudioFile(buf, "w", samplerate, num_channels, bit_depth, format, quality) as f:
                f.write(samples)
            result = buf.getvalue()

        However, this function is much more efficient than the above code, as it writes
        to an in-memory buffer in C++ and avoids interacting with Python at all during the
        encoding process. This allows Python's Global Interpreter Lock (GIL) to be
        released, which also makes this method much more performant in multi-threaded
        programs.

        .. warning::
          This function will encode the entire audio buffer at once, and may consume a
          large amount of memory if the input audio buffer is large.

          To avoid running out of memory with arbitrary-length inputs, it is
          recommended to stream the output into a file or file-like object by using
          :class:`AudioFile` class in write (``"w"``) mode instead.
        """
    pass

class AudioStream:
    """
    A class that allows interacting with live streams of audio from an input
    audio device (i.e.: a microphone, audio interface, etc) and/or to an
    output device (speaker, headphones), allowing access to the audio
    stream from within Python code.

    Use :py:meth:`AudioStream.play` to play audio data to your speakers::

        # Play a 10-second chunk of an audio file:
        with AudioFile("my_audio_file.mp3") as f:
            chunk = f.read(f.samplerate * 10)
        AudioStream.play(chunk, f.samplerate)

    Or use :py:meth:`AudioStream.write` to stream audio in chunks::

        # Play an audio file by looping through it in chunks:
        with AudioStream(output_device="default") as stream:
            with AudioFile("my_audio_file.mp3") as f:
                while f.tell() < f.frames:
                    # Decode and play 512 samples at a time:
                    stream.write(f.read(512))

    :class:`AudioStream` may also be used to pass live audio through a :class:`Pedalboard`::

       # Pass both an input and output device name to connect both ends:
       input_device_name = AudioStream.default_input_device_name
       output_device_name = AudioStream.default_output_device_name
       with AudioStream(input_device_name, output_device_name) as stream:
           # In this block, audio is streaming through `stream`!
           # Audio will be coming out of your speakers at this point.

           # Add plugins to the live audio stream:
           reverb = Reverb()
           stream.plugins.append(reverb)

           # Change plugin properties as the stream is running:
           reverb.wet_level = 1.0

           # Delete plugins:
           del stream.plugins[0]

       # Or use AudioStream synchronously:
       stream = AudioStream(input_device_name, output_device_name)
       stream.plugins.append(Reverb(wet_level=1.0))
       stream.run()  # Run the stream until Ctrl-C is received

    .. warning::
        The :class:`AudioStream` class implements a context manager interface
        to ensure that audio streams are never left "dangling" (i.e.: running in
        the background without being stopped).

        While it is possible to call the :meth:`__enter__` method directly to run an
        audio stream in the background, this can have some nasty side effects. If the
        :class:`AudioStream` object is no longer reachable (not bound to a variable,
        not in scope, etc), the audio stream will continue to run forever, and
        won't stop until the Python interpreter exits.

        To run an :class:`AudioStream` in the background, use Python's
        :py:mod:`threading` module to run the stream object method on a
        background thread, allowing for easier cleanup.

    *Introduced in v0.7.0 for macOS and Windows. Linux support introduced in v0.9.14.*

    :py:meth:`read` *and* :py:meth:`write` *methods introduced in v0.9.12.*
    """

    def __enter__(self) -> AudioStream:
        """
        Use this :class:`AudioStream` as a context manager. Entering the context manager will immediately start the audio stream, sending audio through to the output device.
        """

    def __exit__(self, arg0: object, arg1: object, arg2: object) -> None:
        """
        Exit the context manager, ending the audio stream. Once called, the audio stream will be stopped (i.e.: :py:attr:`running` will be :py:const:`False`).
        """

    def __init__(
        self,
        input_device_name: typing.Optional[str] = None,
        output_device_name: typing.Optional[str] = None,
        plugins: typing.Optional[pedalboard_native.utils.Chain] = None,
        sample_rate: typing.Optional[float] = None,
        buffer_size: typing.Optional[int] = None,
        allow_feedback: bool = False,
        num_input_channels: int = 1,
        num_output_channels: int = 2,
    ) -> None: ...
    def __repr__(self) -> str: ...
    def close(self) -> None:
        """
        Close the audio stream, stopping the audio device and releasing any resources. After calling close, this AudioStream object is no longer usable.
        """

    @staticmethod
    def play(
        audio: NDArray[float32],
        sample_rate: float,
        output_device_name: typing.Optional[str] = None,
    ) -> None:
        """
        Play audio data to the speaker, headphones, or other output device. This method will block until the audio is finished playing.
        """

    def read(self, num_samples: int = 0) -> NDArray[float32]:
        """
        .. warning::
            Recording audio is a **real-time** operation, so if your code doesn't call :py:meth:`read` quickly enough, some audio will be lost. To warn about this, :py:meth:`read` will throw an exception if audio data is dropped. This behavior can be disabled by setting :py:attr:`ignore_dropped_input` to :py:const:`True`. The number of dropped samples since the last call to :py:meth:`read` can be retrieved by accessing the :py:attr:`dropped_input_frame_count` property.
        """

    def run(self) -> None:
        """
        Start streaming audio from input to output, passing the audio stream through the :py:attr:`plugins` on this AudioStream object. This call will block the current thread until a :py:exc:`KeyboardInterrupt` (``Ctrl-C``) is received.
        """

    def write(
        self, audio: NDArray[float32], sample_rate: float
    ) -> None:
        """
        If the provided sample rate does not match the output device's sample rate, an error will be thrown. In this case, you can use :py:class:`StreamResampler` to resample the audio before calling :py:meth:`write`.
        """

    @property
    def buffer_size(self) -> int:
        """
        The size (in frames) of the buffer used between the audio hardware and Python.


        """

    @property
    def buffered_input_sample_count(self) -> typing.Optional[int]:
        """
        The number of frames of audio that are currently in the input buffer. This number will change rapidly, and will be :py:const:`None` if no input device is connected.


        """

    @property
    def dropped_input_frame_count(self) -> int:
        """
        The number of frames of audio that were dropped since the last call to :py:meth:`read`. To prevent audio from being dropped during recording, ensure that you call :py:meth:`read` as often as possible or increase your buffer size.


        """

    @property
    def ignore_dropped_input(self) -> bool:
        """
        Controls the behavior of the :py:meth:`read` method when audio data is dropped. If this property is false (the default), the :py:meth:`read` method will raise a :py:exc:`RuntimeError` if any audio data is dropped in between calls. If this property is true, the :py:meth:`read` method will return the most recent audio data available, even if some audio data was dropped.

        .. note::
            The :py:attr:`dropped_input_frame_count` property is unaffected by this setting.


        """

    @ignore_dropped_input.setter
    def ignore_dropped_input(self, arg1: bool) -> None:
        """
        Controls the behavior of the :py:meth:`read` method when audio data is dropped. If this property is false (the default), the :py:meth:`read` method will raise a :py:exc:`RuntimeError` if any audio data is dropped in between calls. If this property is true, the :py:meth:`read` method will return the most recent audio data available, even if some audio data was dropped.

        .. note::
            The :py:attr:`dropped_input_frame_count` property is unaffected by this setting.
        """

    @property
    def num_input_channels(self) -> int:
        """
        The number of input channels on the input device. Will be ``0`` if no input device is connected.


        """

    @property
    def num_output_channels(self) -> int:
        """
        The number of output channels on the output device. Will be ``0`` if no output device is connected.


        """

    @property
    def plugins(self) -> pedalboard_native.utils.Chain:
        """
        The Pedalboard object that this AudioStream will use to process audio.


        """

    @plugins.setter
    def plugins(self, arg1: pedalboard_native.utils.Chain) -> None:
        """
        The Pedalboard object that this AudioStream will use to process audio.
        """

    @property
    def running(self) -> bool:
        """
        :py:const:`True` if this stream is currently streaming live audio, :py:const:`False` otherwise.


        """

    @property
    def sample_rate(self) -> float:
        """
        The sample rate that this stream is operating at.


        """
    default_input_device_name: Optional[str] = None
    default_output_device_name: Optional[str] = None
    input_device_names: typing.List[str] = []
    output_device_names: typing.List[str] = []
    pass

class ReadableAudioFile(AudioFile):
    """
    A class that wraps an audio file for reading, with native support for Ogg Vorbis,
    MP3, WAV, FLAC, and AIFF files on all operating systems. Other formats may also
    be readable depending on the operating system and installed system libraries:

     - macOS: ``.3g2``, ``.3gp``, ``.aac``, ``.ac3``, ``.adts``, ``.aif``,
       ``.aifc``, ``.aiff``, ``.amr``, ``.au``, ``.bwf``, ``.caf``,
       ``.ec3``, ``.flac``, ``.latm``, ``.loas``, ``.m4a``, ``.m4b``,
       ``.m4r``, ``.mov``, ``.mp1``, ``.mp2``, ``.mp3``, ``.mp4``,
       ``.mpa``, ``.mpeg``, ``.ogg``, ``.qt``, ``.sd2``,
       ``.snd``, ``.w64``, ``.wav``, ``.xhe``
     - Windows: ``.aif``, ``.aiff``, ``.flac``, ``.mp3``, ``.ogg``,
       ``.wav``, ``.wma``
     - Linux: ``.aif``, ``.aiff``, ``.flac``, ``.mp3``, ``.ogg``,
       ``.wav``

    Use :meth:`pedalboard.io.get_supported_read_formats()` to see which
    formats or file extensions are supported on the current platform.

    (Note that although an audio file may have a certain file extension, its
    contents may be encoded with a compression algorithm unsupported by
    Pedalboard.)

    .. note::
        You probably don't want to use this class directly: passing the
        same arguments to :class:`AudioFile` will work too, and allows using
        :class:`AudioFile` just like you'd use ``open(...)`` in Python.
    """

    def __enter__(self) -> ReadableAudioFile:
        """
        Use this :class:`ReadableAudioFile` as a context manager, automatically closing the file and releasing resources when the context manager exits.
        """

    def __exit__(self, arg0: object, arg1: object, arg2: object) -> None:
        """
        Stop using this :class:`ReadableAudioFile` as a context manager, close the file, release its resources.
        """

    @typing.overload
    def __init__(self, filename: str) -> None: ...
    @typing.overload
    def __init__(self, file_like: typing.Union[typing.BinaryIO, memoryview]) -> None: ...

    # These don't exist, but Pyright assumes they do:
    @typing.overload
    def __init__(self, filename: str, mode: Literal["r"]) -> None: ...
    @typing.overload
    def __init__(self, file_like: typing.Union[typing.BinaryIO, memoryview], mode: Literal["r"]) -> None: ...

    @classmethod
    @typing.overload
    def __new__(cls, filename: str) -> ReadableAudioFile: ...
    @classmethod
    @typing.overload
    def __new__(cls, file_like: typing.Union[typing.BinaryIO, memoryview]) -> ReadableAudioFile: ...
    def __repr__(self) -> str: ...
    def close(self) -> None:
        """
        Close this file, rendering this object unusable.
        """

    def read(
        self, num_frames: typing.Union[float, int] = 0
    ) -> NDArray[float32]:
        """
        Read the given number of frames (samples in each channel) from this audio file at its current position.

        ``num_frames`` is a required argument, as audio files can be deceptively large. (Consider that
        an hour-long ``.ogg`` file may be only a handful of megabytes on disk, but may decompress to
        nearly a gigabyte in memory.) Audio files should be read in chunks, rather than all at once, to avoid
        hard-to-debug memory problems and out-of-memory crashes.

        Audio samples are returned as a multi-dimensional :class:`numpy.array` with the shape
        ``(channels, samples)``; i.e.: a stereo audio file will have shape ``(2, <length>)``.
        Returned data is always in the ``float32`` datatype.

        If the file does not contain enough audio data to fill ``num_frames``, the returned
        :class:`numpy.array` will contain as many frames as could be read from the file. (In some cases,
        passing :py:attr:`frames` as ``num_frames`` may still return less data than expected. See documentation
        for :py:attr:`frames` and :py:attr:`exact_duration_known` for more information about situations
        in which this may occur.)

        For most (but not all) audio files, the minimum possible sample value will be ``-1.0f`` and the
        maximum sample value will be ``+1.0f``.

        .. note::
            For convenience, the ``num_frames`` argument may be a floating-point number. However, if the
            provided number of frames contains a fractional part (i.e.: ``1.01`` instead of ``1.00``) then
            an exception will be thrown, as a fractional number of samples cannot be returned.
        """

    def read_raw(
        self, num_frames: typing.Union[float, int] = 0
    ) -> NDArray[typing.Union[np.int8, np.int16, np.int32, np.float32]]:
        """
        Read the given number of frames (samples in each channel) from this audio file at its current position.

        ``num_frames`` is a required argument, as audio files can be deceptively large. (Consider that
        an hour-long ``.ogg`` file may be only a handful of megabytes on disk, but may decompress to
        nearly a gigabyte in memory.) Audio files should be read in chunks, rather than all at once, to avoid
        hard-to-debug memory problems and out-of-memory crashes.

        Audio samples are returned as a multi-dimensional :class:`numpy.array` with the shape
        ``(channels, samples)``; i.e.: a stereo audio file will have shape ``(2, <length>)``.
        Returned data is in the raw format stored by the underlying file (one of ``int8``, ``int16``,
        ``int32``, or ``float32``) and may have any magnitude.

        If the file does not contain enough audio data to fill ``num_frames``, the returned
        :class:`numpy.array` will contain as many frames as could be read from the file. (In some cases,
        passing :py:attr:`frames` as ``num_frames`` may still return less data than expected. See documentation
        for :py:attr:`frames` and :py:attr:`exact_duration_known` for more information about situations
        in which this may occur.)

        .. note::
            For convenience, the ``num_frames`` argument may be a floating-point number. However, if the
            provided number of frames contains a fractional part (i.e.: ``1.01`` instead of ``1.00``) then
            an exception will be thrown, as a fractional number of samples cannot be returned.
        """

    def resampled_to(
        self,
        target_sample_rate: float,
        quality: pedalboard_native.Resample.Quality = pedalboard_native.Resample.Quality.WindowedSinc32,
    ) -> typing.Union[ReadableAudioFile, ResampledReadableAudioFile]:
        """
        Return a :class:`ResampledReadableAudioFile` that will automatically resample this :class:`ReadableAudioFile` to the provided `target_sample_rate`, using a constant amount of memory.

        If `target_sample_rate` matches the existing sample rate of the file, the original file will be returned.

        *Introduced in v0.6.0.*
        """

    def seek(self, position: int) -> None:
        """
        Seek this file to the provided location in frames. Future reads will start from this position.
        """

    def seekable(self) -> bool:
        """
        Returns True if this file is currently open and calls to seek() will work.
        """

    def tell(self) -> int:
        """
        Return the current position of the read pointer in this audio file, in frames. This value will increase as :meth:`read` is called, and may decrease if :meth:`seek` is called.
        """

    @property
    def closed(self) -> bool:
        """
        True iff this file is closed (and no longer usable), False otherwise.


        """

    @property
    def duration(self) -> float:
        """
        The duration of this file in seconds (``frames`` divided by ``samplerate``).

        .. warning::
            :py:attr:`duration` may be an overestimate for certain MP3 files.
            Use :py:attr:`exact_duration_known` property to determine if
            :py:attr:`duration` is accurate. (See the documentation for the
            :py:attr:`frames` attribute for more details.)


        """

    @property
    def exact_duration_known(self) -> bool:
        """
        Returns :py:const:`True` if this file's :py:attr:`frames` and
        :py:attr:`duration` attributes are exact values, or :py:const:`False` if the
        :py:attr:`frames` and :py:attr:`duration` attributes are estimates based
        on the file's size and bitrate.

        If :py:attr:`exact_duration_known` is :py:const:`False`, this value will
        change to :py:const:`True` once the file is read to completion. Once
        :py:const:`True`, this value will not change back to :py:const:`False`
        for the same :py:class:`AudioFile` object (even after calls to :meth:`seek`).

        .. note::
            :py:attr:`exact_duration_known` will only ever be :py:const:`False`
            when reading certain MP3 files. For files in other formats than MP3,
            :py:attr:`exact_duration_known` will always be equal to :py:const:`True`.

        *Introduced in v0.7.2.*


        """

    @property
    def file_dtype(self) -> str:
        """
        The data type (``"int16"``, ``"float32"``, etc) stored natively by this file.

        Note that :meth:`read` will always return a ``float32`` array, regardless of the value of this property. Use :meth:`read_raw` to read data from the file in its ``file_dtype``.


        """

    @property
    def frames(self) -> int:
        """
        The total number of frames (samples per channel) in this file.

        For example, if this file contains 10 seconds of stereo audio at sample rate
        of 44,100 Hz, ``frames`` will return ``441,000``.

        .. warning::
            When reading certain MP3 files that have been encoded in constant bitrate mode,
            the :py:attr:`frames` and :py:attr:`duration` properties may initially be estimates
            and **may change as the file is read**. The :py:attr:`exact_duration_known`
            property indicates if the values of :py:attr:`frames` and :py:attr:`duration`
            are estimates or exact values.

            This discrepancy is due to the fact that MP3 files are not required to have
            headers that indicate the duration of the file. If an MP3 file is opened and a
            ``Xing`` or ``Info`` header frame is not found, the initial value of the
            :py:attr:`frames` and :py:attr:`duration` attributes are estimates based on the file's
            bitrate and size. This may result in an overestimate of the file's duration
            if there is additional data present in the file after the audio stream is finished.

            If the exact number of frames in the file is required, read the entire file
            first before accessing the :py:attr:`frames` or :py:attr:`duration` properties.
            This operation forces each frame to be parsed and guarantees that
            :py:attr:`frames` and :py:attr:`duration` are correct, at the expense of
            scanning the entire file::

                with AudioFile("my_file.mp3") as f:
                    while f.tell() < f.frames:
                        f.read(f.samplerate * 60)

                    # f.frames is now guaranteed to be exact, as the entire file has been read:
                    assert f.exact_duration_known == True

                    f.seek(0)
                    num_channels, num_samples = f.read(f.frames).shape
                    assert num_samples == f.frames

            This behaviour is present in v0.7.2 and later; prior versions would
            raise an exception when trying to read the ends of MP3 files that contained
            trailing non-audio data and lacked ``Xing`` or ``Info`` headers.


        """

    @property
    def name(self) -> typing.Optional[str]:
        """
        The name of this file.

        If this :class:`ReadableAudioFile` was opened from a file-like object, this will be ``None``.


        """

    @property
    def num_channels(self) -> int:
        """
        The number of channels in this file.


        """

    @property
    def samplerate(self) -> typing.Union[float, int]:
        """
        The sample rate of this file in samples (per channel) per second (Hz). Sample rates are represented as floating-point numbers by default, but this property will be an integer if the file's sample rate has no fractional part.


        """
    pass

class ResampledReadableAudioFile(AudioFile):
    """
    A class that wraps an audio file for reading, while resampling
    the audio stream on-the-fly to a new sample rate.

    *Introduced in v0.6.0.*

    Reading, seeking, and all other basic file I/O operations are supported (except for
    :meth:`read_raw`).

    :class:`ResampledReadableAudioFile` should usually
    be used via the :meth:`resampled_to` method on :class:`ReadableAudioFile`:

    ::

       with AudioFile("my_file.mp3").resampled_to(22_050) as f:
           f.samplerate # => 22050
           first_ten_seconds = f.read(int(f.samplerate * 10))

    Fractional (real-valued, non-integer) sample rates are supported.

    Under the hood, :class:`ResampledReadableAudioFile` uses a stateful
    :class:`StreamResampler` instance, which uses a constant amount of
    memory to resample potentially-unbounded streams of audio. The audio
    output by :class:`ResampledReadableAudioFile` will always be identical
    to the result obtained by passing the entire audio file through a
    :class:`StreamResampler`, with the added benefits of allowing chunked
    reads, seeking through files, and using a constant amount of memory.
    """

    def __enter__(self) -> ResampledReadableAudioFile:
        """
        Use this :class:`ResampledReadableAudioFile` as a context manager, automatically closing the file and releasing resources when the context manager exits.
        """

    def __exit__(self, arg0: object, arg1: object, arg2: object) -> None:
        """
        Stop using this :class:`ResampledReadableAudioFile` as a context manager, close the file, release its resources.
        """

    def __init__(
        self,
        audio_file: ReadableAudioFile,
        target_sample_rate: float,
        resampling_quality: pedalboard_native.Resample.Quality = pedalboard_native.Resample.Quality.WindowedSinc32,
    ) -> None: ...
    @classmethod
    def __new__(
        cls,
        audio_file: ReadableAudioFile,
        target_sample_rate: float,
        resampling_quality: pedalboard_native.Resample.Quality = pedalboard_native.Resample.Quality.WindowedSinc32,
    ) -> ResampledReadableAudioFile: ...
    def __repr__(self) -> str: ...
    def close(self) -> None:
        """
        Close this file, rendering this object unusable. Note that the :class:`ReadableAudioFile` instance that is wrapped by this object will not be closed, and will remain usable.
        """

    def read(
        self, num_frames: typing.Union[float, int] = 0
    ) -> NDArray[float32]:
        """
        Read the given number of frames (samples in each channel, at the target sample rate)
        from this audio file at its current position, automatically resampling on-the-fly to
        ``target_sample_rate``.

        ``num_frames`` is a required argument, as audio files can be deceptively large. (Consider that
        an hour-long ``.ogg`` file may be only a handful of megabytes on disk, but may decompress to
        nearly a gigabyte in memory.) Audio files should be read in chunks, rather than all at once, to avoid
        hard-to-debug memory problems and out-of-memory crashes.

        Audio samples are returned as a multi-dimensional :class:`numpy.array` with the shape
        ``(channels, samples)``; i.e.: a stereo audio file will have shape ``(2, <length>)``.
        Returned data is always in the ``float32`` datatype.

        If the file does not contain enough audio data to fill ``num_frames``, the returned
        :class:`numpy.array` will contain as many frames as could be read from the file. (In some cases,
        passing :py:attr:`frames` as ``num_frames`` may still return less data than expected. See documentation
        for :py:attr:`frames` and :py:attr:`exact_duration_known` for more information about situations
        in which this may occur.)

        For most (but not all) audio files, the minimum possible sample value will be ``-1.0f`` and the
        maximum sample value will be ``+1.0f``.

        .. note::
            For convenience, the ``num_frames`` argument may be a floating-point number. However, if the
            provided number of frames contains a fractional part (i.e.: ``1.01`` instead of ``1.00``) then
            an exception will be thrown, as a fractional number of samples cannot be returned.
        """

    def seek(self, position: int) -> None:
        """
        Seek this file to the provided location in frames at the target sample rate. Future reads will start from this position.

        .. note::
            Prior to version 0.7.3, this method operated in linear time with respect to the seek position (i.e.: the file was seeked to its beginning and pushed through the resampler) to ensure that the resampled audio output was sample-accurate. This was optimized in version 0.7.3 to operate in effectively constant time while retaining sample-accuracy.
        """

    def seekable(self) -> bool:
        """
        Returns True if this file is currently open and calls to seek() will work.
        """

    def tell(self) -> int:
        """
        Return the current position of the read pointer in this audio file, in frames at the target sample rate. This value will increase as :meth:`read` is called, and may decrease if :meth:`seek` is called.
        """

    @property
    def closed(self) -> bool:
        """
        True iff either this file or its wrapped :class:`ReadableAudioFile` instance are closed (and no longer usable), False otherwise.


        """

    @property
    def duration(self) -> float:
        """
        The duration of this file in seconds (``frames`` divided by ``samplerate``).

        .. warning::
            When reading certain MP3 files, the :py:attr:`frames` and :py:attr:`duration` properties may initially be estimates and **may change as the file is read**. See the documentation for :py:attr:`.ReadableAudioFile.frames` for more details.


        """

    @property
    def exact_duration_known(self) -> bool:
        """
        Returns :py:const:`True` if this file's :py:attr:`frames` and
        :py:attr:`duration` attributes are exact values, or :py:const:`False` if the
        :py:attr:`frames` and :py:attr:`duration` attributes are estimates based
        on the file's size and bitrate.

        :py:attr:`exact_duration_known` will change from :py:const:`False` to
        :py:const:`True` as the file is read to completion. Once :py:const:`True`,
        this value will not change back to :py:const:`False` for the same
        :py:class:`AudioFile` object (even after calls to :meth:`seek`).

        .. note::
            :py:attr:`exact_duration_known` will only ever be :py:const:`False`
            when reading certain MP3 files. For files in other formats than MP3,
            :py:attr:`exact_duration_known` will always be equal to :py:const:`True`.

        *Introduced in v0.7.2.*


        """

    @property
    def file_dtype(self) -> str:
        """
        The data type (``"int16"``, ``"float32"``, etc) stored natively by this file.

        Note that :meth:`read` will always return a ``float32`` array, regardless of the value of this property.


        """

    @property
    def frames(self) -> int:
        """
        The total number of frames (samples per channel) in this file, at the target sample rate.

        For example, if this file contains 10 seconds of stereo audio at sample rate of 44,100 Hz, and ``target_sample_rate`` is 22,050 Hz, ``frames`` will return ``22,050``.

        Note that different ``resampling_quality`` values used for resampling may cause ``frames`` to differ by ± 1 from its expected value.

        .. warning::
            When reading certain MP3 files, the :py:attr:`frames` and :py:attr:`duration` properties may initially be estimates and **may change as the file is read**. See the documentation for :py:attr:`.ReadableAudioFile.frames` for more details.


        """

    @property
    def name(self) -> typing.Optional[str]:
        """
        The name of this file.

        If the :class:`ReadableAudioFile` wrapped by this :class:`ResampledReadableAudioFile` was opened from a file-like object, this will be ``None``.


        """

    @property
    def num_channels(self) -> int:
        """
        The number of channels in this file.


        """

    @property
    def resampling_quality(self) -> pedalboard_native.Resample.Quality:
        """
        The resampling algorithm used to resample from the original file's sample rate to the ``target_sample_rate``.


        """

    @property
    def samplerate(self) -> typing.Union[float, int]:
        """
        The sample rate of this file in samples (per channel) per second (Hz). This will be equal to the ``target_sample_rate`` parameter passed when this object was created. Sample rates are represented as floating-point numbers by default, but this property will be an integer if the file's target sample rate has no fractional part.


        """
    pass

class StreamResampler:
    """
    A streaming resampler that can change the sample rate of multiple chunks of audio in series, while using constant memory.

    For a resampling plug-in that can be used in :class:`Pedalboard` objects, see :class:`pedalboard.Resample`.

    *Introduced in v0.6.0.*
    """

    def __init__(
        self,
        source_sample_rate: float,
        target_sample_rate: float,
        num_channels: int,
        quality: pedalboard_native.Resample.Quality = pedalboard_native.Resample.Quality.WindowedSinc32,
    ) -> None:
        """
        Create a new StreamResampler, capable of resampling a potentially-unbounded audio stream with a constant amount of memory. The source sample rate, target sample rate, quality, or number of channels cannot be changed once the resampler is instantiated.
        """

    def __repr__(self) -> str: ...
    def process(
        self, input: typing.Optional[NDArray[float32]] = None
    ) -> NDArray[float32]:
        """
        Resample a 32-bit floating-point audio buffer. The returned buffer may be smaller than the provided buffer depending on the quality method used. Call :meth:`process()` without any arguments to flush the internal buffers and return all remaining audio.
        """

    def reset(self) -> None:
        """
        Used to reset the internal state of this resampler. Call this method when resampling a new audio stream to prevent audio from leaking between streams.
        """

    @property
    def input_latency(self) -> float:
        """
        The number of samples (in the input sample rate) that must be supplied before this resampler will begin returning output.


        """

    @property
    def num_channels(self) -> int:
        """
        The number of channels expected to be passed in every call to :meth:`process()`.


        """

    @property
    def quality(self) -> pedalboard_native.Resample.Quality:
        """
        The resampling algorithm used by this resampler.


        """

    @property
    def source_sample_rate(self) -> float:
        """
        The source sample rate of the input audio that this resampler expects to be passed to :meth:`process()`.


        """

    @property
    def target_sample_rate(self) -> float:
        """
        The sample rate of the audio that this resampler will return from :meth:`process()`.


        """
    pass

class WriteableAudioFile(AudioFile):
    """
    A class that wraps an audio file for writing, with native support for Ogg Vorbis,
    MP3, WAV, FLAC, and AIFF files on all operating systems.

    Use :meth:`pedalboard.io.get_supported_write_formats()` to see which
    formats or file extensions are supported on the current platform.

    Args:
        filename_or_file_like:
            The path to an output file to write to, or a seekable file-like
            binary object (like ``io.BytesIO``) to write to.

        samplerate:
            The sample rate of the audio that will be written to this file.
            All calls to the :meth:`write` method will assume this sample rate
            is used.

        num_channels:
            The number of channels in the audio that will be written to this file.
            All calls to the :meth:`write` method will expect audio with this many
            channels, and will throw an exception if the audio does not contain
            this number of channels.

        bit_depth:
            The bit depth (number of bits per sample) that will be written
            to this file. Used for raw formats like WAV and AIFF. Will have no effect
            on compressed formats like MP3 or Ogg Vorbis.

        quality:
            An optional string or number that indicates the quality level to use
            for the given audio compression codec. Different codecs have different
            compression quality values; numeric values like ``128`` and ``256`` will
            usually indicate the number of kilobits per second used by the codec.
            Some formats, like MP3, support more advanced options like ``V2`` (as
            specified by `the LAME encoder <https://lame.sourceforge.io/>`_) which
            may be passed as a string. The strings ``"best"``, ``"worst"``,
            ``"fastest"``, and ``"slowest"`` will also work for any codec.

    .. note::
        You probably don't want to use this class directly: all of the parameters
        accepted by the :class:`WriteableAudioFile` constructor will be accepted by
        :class:`AudioFile` as well, as long as the ``"w"`` mode is passed as the
        second argument.
    """

    def __enter__(self) -> WriteableAudioFile: ...
    def __exit__(self, arg0: object, arg1: object, arg2: object) -> None: ...
    @typing.overload
    def __init__(
        self,
        filename: str,
        samplerate: float,
        num_channels: int = 1,
        bit_depth: int = 16,
        quality: typing.Optional[typing.Union[str, float]] = None,
        codec_options: typing.Dict[WriteableAudioFileFlag, typing.Union[bool, float, int, str]] = ...,
    ) -> None: ...
    @typing.overload
    def __init__(
        self,
        file_like: typing.BinaryIO,
        samplerate: float,
        num_channels: int = 1,
        bit_depth: int = 16,
        quality: typing.Optional[typing.Union[str, float]] = None,
        format: typing.Optional[str] = None,
        codec_options: typing.Dict[WriteableAudioFileFlag, typing.Union[bool, float, int, str]] = ...,
    ) -> None: ...
    @classmethod
    @typing.overload
    def __new__(
        cls,
        filename: str,
        samplerate: typing.Optional[float] = None,
        num_channels: int = 1,
        bit_depth: int = 16,
        quality: typing.Optional[typing.Union[str, float]] = None,
        codec_options: typing.Dict[WriteableAudioFileFlag, typing.Union[bool, float, int, str]] = ...,
    ) -> WriteableAudioFile: ...
    @classmethod
    @typing.overload
    def __new__(
        cls,
        file_like: typing.BinaryIO,
        samplerate: typing.Optional[float] = None,
        num_channels: int = 1,
        bit_depth: int = 16,
        quality: typing.Optional[typing.Union[str, float]] = None,
        format: typing.Optional[str] = None,
        codec_options: typing.Dict[WriteableAudioFileFlag, typing.Union[bool, float, int, str]] = ...,
    ) -> WriteableAudioFile: ...

    # This overload does not actually exist; just makes Pyright happy as
    # it assumes that __init__ is called with the same arguments as __new__:
    @typing.overload
    def __init__(
        self,
        filename: str,
        mode: Literal["w"],
        samplerate: float,
        num_channels: int = 1,
        bit_depth: int = 16,
        quality: typing.Optional[typing.Union[str, float]] = None,
        codec_options: typing.Dict[WriteableAudioFileFlag, typing.Union[bool, float, int, str]] = ...,
    ) -> None: ...

    # This overload does not actually exist; just makes Pyright happy as
    # it assumes that __init__ is called with the same arguments as __new__:
    @typing.overload
    def __init__(
        self,
        file_like: typing.BinaryIO,
        mode: Literal["w"],
        samplerate: typing.Optional[float] = None,
        num_channels: int = 1,
        bit_depth: int = 16,
        quality: typing.Optional[typing.Union[str, float]] = None,
        format: typing.Optional[str] = None,
        codec_options: typing.Dict[WriteableAudioFileFlag, typing.Union[bool, float, int, str]] = ...,
    ) -> None: ...

    def __repr__(self) -> str: ...
    def close(self) -> None:
        """
        Close this file, flushing its contents to disk and rendering this object unusable for further writing.
        """

    def flush(self) -> None:
        """
        Attempt to flush this audio file's contents to disk. Not all formats support flushing, so this may throw a RuntimeError. (If this happens, closing the file will reliably force a flush to occur.)
        """

    def tell(self) -> int:
        """
        Return the current position of the write pointer in this audio file, in frames at the target sample rate. This value will increase as :meth:`write` is called, and will never decrease.
        """

    def write(self, samples: numpy.ndarray) -> None:
        """
        Encode an array of audio data and write it to this file. The number of channels in the array must match the number of channels used to open the file. The array may contain audio in any shape. If the file's bit depth or format does not match the provided data type, the audio will be automatically converted.

        Arrays of type int8, int16, int32, float32, and float64 are supported. If an array of an unsupported ``dtype`` is provided, a ``TypeError`` will be raised.

        .. warning::
            If an array of shape ``(num_channels, num_channels)`` is passed to this method before any other audio data is provided, an exception will be thrown, as the method will not be able to infer which dimension of the input corresponds to the number of channels and which dimension corresponds to the number of samples.

            To avoid this, first call this method with an array where the number of samples does not match the number of channels.

            The channel layout from the most recently provided input will be cached on the :py:class:`WritableAudioFile` object and will be used if necessary to disambiguate the array layout:

            .. code-block:: python

                with AudioFile("my_file.mp3", "w", 44100, num_channels=2) as f:
                    # This will throw an exception:
                    f.write(np.zeros((2, 2)))
                    # But this will work:
                    f.write(np.zeros((2, 1)))
                    # And now `f` expects an input shape of (num_channels, num_samples), so this works:
                    f.write(np.zeros((2, 2)))

                # Also an option: pass (0, num_channels) or (num_channels, 0) first
                # to hint that the input will be in that shape without writing anything:
                with AudioFile("my_file.mp3", "w", 44100, num_channels=2) as f:
                    # Pass a hint, but write nothing:
                    f.write(np.zeros((2, 0)))
                    # And now `f` expects an input shape of (num_channels, num_samples), so this works:
                    f.write(np.zeros((2, 2)))
        """

    @property
    def closed(self) -> bool:
        """
        If this file has been closed, this property will be True.


        """

    @property
    def file_dtype(self) -> str:
        """
        The data type stored natively by this file. Note that write(...) will accept multiple datatypes, regardless of the value of this property.


        """

    @property
    def frames(self) -> int:
        """
        The total number of frames (samples per channel) written to this file so far.


        """

    @property
    def num_channels(self) -> int:
        """
        The number of channels in this file.


        """

    @property
    def quality(self) -> typing.Optional[str]:
        """
        The quality setting used to write this file. For many formats, this may be ``None``.

        Quality options differ based on the audio codec used in the file. Most codecs specify a number of bits per second in 16- or 32-bit-per-second increments (128 kbps, 160 kbps, etc). Some codecs provide string-like options for variable bit-rate encoding (i.e. "V0" through "V9" for MP3). The strings ``"best"``, ``"worst"``, ``"fastest"``, and ``"slowest"`` will also work for any codec.


        """

    @property
    def samplerate(self) -> typing.Union[float, int]:
        """
        The sample rate of this file in samples (per channel) per second (Hz). Sample rates are represented as floating-point numbers by default, but this property will be an integer if the file's sample rate has no fractional part.


        """
    pass

def get_supported_read_formats() -> typing.List[str]:
    pass

def get_supported_write_formats() -> typing.List[str]:
    pass



#! /usr/bin/env python
#
# Copyright 2023 Spotify AB
#
# Licensed under the GNU Public License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.gnu.org/licenses/gpl-3.0.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Counter
from typing import List, Tuple, Union


def parse_midi_message_part(byte: Union[int, str, bytes]) -> int:
    if isinstance(byte, int):
        return byte
    elif isinstance(byte, bytes):
        if len(byte) == 1:
            return byte[0]
        else:
            raise NotImplementedError("Not sure how to interpret provided MIDI message.")
    elif isinstance(byte, str):
        try:
            return int(byte)
        except ValueError as e:
            raise NotImplementedError("Not sure how to interpret provided MIDI message.") from e
    raise NotImplementedError("MIDI messages must currently be bytes or lists of byte values.")


def parse_midi_message_string(midi_message_string: str) -> bytes:
    raise NotImplementedError("MIDI messages must currently be bytes or lists of byte values.")


def normalize_midi_messages(_input) -> List[Tuple[bytes, float]]:
    """
    Given a duck-typed Python input, usually an iterable of MIDI messages,
    normalize the input to a list of tuples of bytes which can be converted
    into a juce::MidiBuffer on the C++ side.
    """
    output = []
    for message in _input:
        if hasattr(message, "bytes") and hasattr(message, "time"):
            output.append((bytes(message.bytes()), message.time))
        elif (isinstance(message, tuple) or isinstance(message, list)) and len(message) == 2:
            message, time = message
            if isinstance(message, str):
                message = parse_midi_message_string(message)
            if isinstance(message, list):
                message = bytes([parse_midi_message_part(x) for x in message])
            elif not isinstance(message, bytes):
                message = bytes(message)
            output.append((message, time))

    # Detect the case in which the provided timestamps
    # are likely delta values rather than absolute values:
    all_timestamps = [t for _, t in output]
    if len(all_timestamps) > 100 and len(set(all_timestamps)) > 1:
        all_timestamps_histogram = Counter(all_timestamps)
        (
            most_common_timestamp,
            num_instances_of_most_common,
        ) = all_timestamps_histogram.most_common()[0]
        if num_instances_of_most_common > 100:
            raise ValueError(
                "Pedalboard requires MIDI input timestamps to be absolute values, specified "
                "as the number of seconds from the start of the returned audio buffer. The "
                f"provided MIDI data contains {num_instances_of_most_common:,} events at timestamp "
                f"{most_common_timestamp}, which suggests that the timestamps may be delta values "
                "rather than absolute values. Try converting your MIDI message timestamps to "
                "absolute values first."
            )

    return output



# In certain situations (including when using asyncio), Pedalboard's C++ code will
# implicitly and lazily import NumPy via its Pybind11 bindings. This lazy import
# can sometimes (!) cause a deadlock on the GIL if multiple threads are attempting
# it simultaneously. To work around this, we preload Numpy into the current Python
# process so that Pybind11's lazy import is effectively a no-op.
import numpy  # noqa
from pedalboard_native.io import *  # noqa: F403, F401 # type: ignore



from __future__ import annotations
import pedalboard_native._internal

import typing

original_overload = typing.overload
__OVERLOADED_DOCSTRINGS = {}

def patch_overload(func):
    original_overload(func)
    if func.__doc__:
        __OVERLOADED_DOCSTRINGS[func.__qualname__] = func.__doc__
    else:
        func.__doc__ = __OVERLOADED_DOCSTRINGS.get(func.__qualname__)
    if func.__doc__:
        # Work around the fact that pybind11-stubgen generates
        # duplicate docstrings sometimes, once for each overload:
        docstring = func.__doc__
        if docstring[len(docstring) // 2 :].strip() == docstring[: -len(docstring) // 2].strip():
            func.__doc__ = docstring[len(docstring) // 2 :].strip()
    return func

typing.overload = patch_overload

from typing_extensions import Literal
from enum import Enum
import threading
import pedalboard_native

__all__ = [
    "AddLatency",
    "FixedSizeBlockTestPlugin",
    "ForceMonoTestPlugin",
    "PrimeWithSilenceTestPlugin",
    "ResampleWithLatency",
]

class AddLatency(pedalboard_native.Plugin):
    """
    A dummy plugin that delays input audio for the given number of samples before passing it back to the output. Used internally to test Pedalboard's automatic latency compensation. Probably not useful as a real effect.
    """

    def __init__(self, samples: int = 44100) -> None: ...
    pass

class FixedSizeBlockTestPlugin(pedalboard_native.Plugin):
    def __init__(self, expected_block_size: int = 160) -> None: ...
    def __repr__(self) -> str: ...
    pass

class ForceMonoTestPlugin(pedalboard_native.Plugin):
    def __repr__(self) -> str: ...
    pass

class PrimeWithSilenceTestPlugin(pedalboard_native.Plugin):
    def __init__(self, expected_silent_samples: int = 160) -> None: ...
    def __repr__(self) -> str: ...
    pass

class ResampleWithLatency(pedalboard_native.Plugin):
    def __init__(
        self,
        target_sample_rate: float = 8000.0,
        internal_latency: int = 1024,
        quality: pedalboard_native.Resample.Quality = pedalboard_native.Resample.Quality.WindowedSinc32,
    ) -> None: ...
    def __repr__(self) -> str: ...
    @property
    def quality(self) -> pedalboard_native.Resample.Quality:
        """ """

    @quality.setter
    def quality(self, arg1: pedalboard_native.Resample.Quality) -> None:
        pass

    @property
    def target_sample_rate(self) -> float:
        """ """

    @target_sample_rate.setter
    def target_sample_rate(self, arg1: float) -> None:
        pass
    pass



#! /usr/bin/env python
#
# Copyright 2021 Spotify AB
#
# Licensed under the GNU Public License, Version 3.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.gnu.org/licenses/gpl-3.0.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform
import re
import weakref
from contextlib import contextmanager
from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
    no_type_check,
)

from pedalboard_native import (  # type: ignore
    ExternalPlugin,  # type: ignore
    Plugin,
    _AudioProcessorParameter,
)
from pedalboard_native.utils import Chain  # type: ignore


class Pedalboard(Chain):
    """
    A container for a series of :class:`Plugin` objects, to use for processing audio, like a
    `guitar pedalboard <https://en.wikipedia.org/wiki/Guitar_pedalboard>`_.

    :class:`Pedalboard` objects act like regular Python ``List`` objects,
    but come with an additional :py:meth:`process` method (also aliased to :py:meth:`__call__`),
    allowing audio to be passed through the entire :class:`Pedalboard` object for processing::

        my_pedalboard = Pedalboard()
        my_pedalboard.append(Reverb())
        output_audio = my_pedalboard(input_audio)

    .. warning::
        :class:`Pedalboard` objects may only contain effects plugins (i.e.: those for which
        :attr:`is_effect` is ``True``), and cannot contain instrument plugins (i.e.: those
        for which :attr:`is_instrument` is ``True``).
    """

    def __init__(self, plugins: Optional[List[Plugin]] = None):
        super().__init__(plugins or [])

    def __repr__(self) -> str:
        return "<{} with {} plugin{}: {}>".format(
            self.__class__.__name__,
            len(self),
            "" if len(self) == 1 else "s",
            list(self),
        )


# Stored in a trie to speed up these matches, as they happen millions of times in test:
FLOAT_SUFFIXES_TO_IGNORE = ["x", "%", "*", ",", ".", "hz", "ms", "db", "sec", "dbtp", "seconds"]


class _SuffixTrie:
    def __init__(self, char: Optional[str]):
        self.char = char
        self.children: Dict[str, Optional["_SuffixTrie"]] = {}
        self.is_match = False

    def __repr__(self) -> str:
        return f"<_SuffixTrie {self.char} {self.is_match} {self.children}>"


FLOAT_SUFFIX_TRIE = _SuffixTrie("")
for suffix in FLOAT_SUFFIXES_TO_IGNORE:
    node = FLOAT_SUFFIX_TRIE
    for char in suffix[::-1]:
        if char not in node.children:
            node.children[char] = _SuffixTrie(char)
        node = cast(_SuffixTrie, node.children[char])
    node.is_match = True


def strip_common_float_suffixes(
    s: Union[float, str, bool], strip_si_prefixes: bool = True
) -> Union[float, str, bool]:
    if not isinstance(s, str) or (hasattr(s, "type") and s.type is not str):  # type: ignore
        return s

    s = s.strip()
    s_lower = s.lower()

    # Handle certain plugins that report "Hz" and "kHz" suffixes:
    if strip_si_prefixes and s_lower.endswith("khz") and len(s) >= 3:
        try:
            return str(float(s[:-3]) * 1000)
        except ValueError:
            return s

    while True:
        node = FLOAT_SUFFIX_TRIE
        for i in reversed(range(len(s_lower))):
            char = s_lower[i]
            if char in node.children:
                node = cast(_SuffixTrie, node.children[char])
                if node.is_match:
                    s = s[:i].strip()
                    s_lower = s_lower[: len(s)]
                    break
            else:
                return s
        else:
            return s


def looks_like_float(s: Union[float, str]) -> bool:
    if isinstance(s, float):
        return True

    try:
        float(strip_common_float_suffixes(s))
        return True
    except ValueError:
        return False


class ReadOnlyDictWrapper(dict):
    def __setitem__(self, name, value):
        raise TypeError(
            "The .parameters dictionary on a Plugin instance returns "
            "a read-only dictionary of its parameters. To change a "
            "parameter, set the parameter on the plugin directly as "
            f"an attribute. (`my_plugin.{name} = {value}`)"
        )


# This is so nasty that I had to disable MyPy. Sorry.
@no_type_check
def wrap_type(base_type):
    @no_type_check
    class WeakTypeWrapper(base_type):
        """
        A wrapper around `base_type` that allows adding additional
        accessors through a weak reference. Useful for syntax convenience.
        """

        def __new__(cls, value, *args, **kwargs):
            try:
                return base_type.__new__(cls, value)
            except TypeError:
                return base_type.__new__(cls)

        def __init__(self, *args, **kwargs):
            if "wrapped" in kwargs:
                self._wrapped = weakref.ref(kwargs["wrapped"])
                del kwargs["wrapped"]
            else:
                raise ValueError(
                    "WeakTypeWrapper({}) expected to be passed a 'wrapped' kwarg.".format(base_type)
                )
            try:
                super().__init__(*args, **kwargs)
            except TypeError:
                pass

        def __getattr__(self, name):
            wrapped = self._wrapped()
            if hasattr(wrapped, name):
                return getattr(wrapped, name)
            if hasattr(super(), "__getattr__"):
                return super().__getattr__(name)
            raise AttributeError("'{}' has no attribute '{}'".format(base_type.__name__, name))

        def __dir__(self) -> Iterable[str]:
            wrapped = self._wrapped()
            if wrapped:
                return list(dir(wrapped)) + list(super().__dir__())
            return super().__dir__()

    return WeakTypeWrapper


@no_type_check
class WrappedBool(object):
    def __init__(self, value):
        if not isinstance(value, bool):
            raise TypeError(f"WrappedBool should be passed a boolean, got {type(value)}")
        self.__value = value

    def __repr__(self):
        return repr(self.__value)

    def __eq__(self, o: object) -> bool:
        return self.__value == o

    def __hash__(self) -> int:
        return hash(self.__value)

    def __bool__(self):
        return bool(self.__value)

    def __str__(self):
        return str(self.__value)

    def __getattr__(self, attr: str):
        return getattr(self.__value, attr)

    def __hasattr__(self, attr: str):
        return hasattr(self.__value, attr)


StringWithParameter: Type[str] = wrap_type(str)
FloatWithParameter: Type[float] = wrap_type(float)
BooleanWithParameter: Type[bool] = wrap_type(WrappedBool)


# Some plugins, on some platforms (TAL Reverb 3 on Ubuntu, so far) seem to
# expose every possible MIDI CC value as a separate parameter
# (i.e.: MIDI CC [0-16]|[0-128], resulting in 2,048 parameters).
# This hugely delays load times and adds complexity to the interface.
# Guitar Rig also seems to expose 512 parameters, each matching "P\d\d\d"
PARAMETER_NAME_REGEXES_TO_IGNORE = set(
    [re.compile(pattern) for pattern in ["MIDI CC ", r"P\d\d\d"]]
)

TRUE_BOOLEANS: Set[str] = {"on", "yes", "true", "enabled"}


def get_text_for_raw_value(
    cpp_parameter: _AudioProcessorParameter, raw_value: float, slow: bool = False
) -> Optional[str]:
    # Some plugins don't respond properly to get_text_for_raw_value,
    # but do respond when a parameter's raw value is changed.
    # This helper method works around this issue.
    if slow:
        original_value = cpp_parameter.raw_value
        try:
            cpp_parameter.raw_value = raw_value
            return cpp_parameter.string_value
        finally:
            cpp_parameter.raw_value = original_value
    else:
        return cpp_parameter.get_text_for_raw_value(raw_value)


class AudioProcessorParameter(object):
    """
    A wrapper around various different parameters exposed by
    :class:`VST3Plugin` or :class:`AudioUnitPlugin` instances.

    :class:`AudioProcessorParameter` objects are rarely used directly,
    and usually used via their implicit interface::

       my_plugin = load_plugin("My Cool Audio Effect.vst3")
       # Print all of the parameter names:
       print(my_plugin.parameters.keys())
       # ["mix", "delay_time_ms", "foobar"]
       # Access each parameter as if it were just a Python attribute:
       my_plugin.mix = 0.5
       my_plugin.delay_time_ms = 400

    .. note::
        :class:`AudioProcessorParameter` tries to guess the range of
        valid parameter values, as well as the type/unit of the parameter,
        when instantiated. This guess may not always be accurate.
        Raw control over the underlying parameter's value can be had by
        accessing the :py:attr:`raw_value` attribute, which is always bounded
        on [0, 1] and is passed directly to the underlying plugin object.
    """

    def __init__(self, plugin: "ExternalPlugin", parameter_name: str, search_steps: int = 1000):
        self.__plugin = plugin
        self.__parameter_name = parameter_name

        self.ranges: Dict[Tuple[float, float], Union[str, float, bool]] = {}

        with self.__get_cpp_parameter() as cpp_parameter:
            text_value: Optional[str] = None
            start_of_range: float = 0

            for fetch_slow in (False, True):
                start_of_range: float = 0
                text_value: Optional[str] = None

                self.ranges = {}
                for x in range(0, search_steps + 1):
                    raw_value = x / search_steps
                    x_text_value = get_text_for_raw_value(cpp_parameter, raw_value, fetch_slow)
                    if text_value is None:
                        text_value = x_text_value
                    elif x_text_value != text_value:
                        # End current range and start a new one
                        self.ranges[(start_of_range, raw_value)] = text_value
                        text_value = x_text_value
                        start_of_range = raw_value
                results_look_incorrect = not self.ranges or (
                    len(self.ranges) == 1 and all(looks_like_float(v) for v in self.ranges.values())
                )
                if not results_look_incorrect:
                    break
            if text_value is None:
                raise NotImplementedError(
                    f"Plugin parameter '{parameter_name}' failed to return a valid string for its"
                    " value."
                )
            self.ranges[(start_of_range, 1)] = text_value

            self.python_name = to_python_parameter_name(cpp_parameter)

        self.min_value = None
        self.max_value = None
        self.step_size = None
        self.approximate_step_size = None
        self.type: Type = str

        if all(looks_like_float(v) for v in self.ranges.values()):
            self.type = float
            float_ranges = {
                k: float(strip_common_float_suffixes(v)) for k, v in self.ranges.items()
            }
            self.min_value = min(float_ranges.values())
            self.max_value = max(float_ranges.values())

            if not self.label:
                # Try to infer the label from the string values:
                a_value = next(iter(self.ranges.values()))
                if isinstance(a_value, str):
                    stripped_value = strip_common_float_suffixes(a_value, strip_si_prefixes=False)
                    if (
                        stripped_value != a_value
                        and isinstance(stripped_value, str)
                        and stripped_value in a_value
                    ):
                        # We probably have a label! Iterate through all values just to make sure:
                        all_possible_labels = set()
                        for value in self.ranges.values():
                            if not isinstance(value, str):
                                continue
                            stripped_value = strip_common_float_suffixes(
                                value, strip_si_prefixes=False
                            )
                            if not isinstance(stripped_value, str):
                                continue
                            all_possible_labels.add(value.replace(stripped_value, "").strip())
                        if len(all_possible_labels) == 1:
                            self._label = next(iter(all_possible_labels))

            sorted_values = sorted(float_ranges.values())
            first_derivative_steps = set(
                [round(abs(b - a), 8) for a, b in zip(sorted_values, sorted_values[1:])]
            )
            if len(first_derivative_steps) == 1:
                self.step_size = next(iter(first_derivative_steps))
            elif first_derivative_steps:
                self.approximate_step_size = sum(first_derivative_steps) / len(
                    first_derivative_steps
                )
            self.ranges = dict(float_ranges)
        elif len(self.ranges) == 2 and (
            TRUE_BOOLEANS & {(v.lower() if isinstance(v, str) else v) for v in self.ranges.values()}
        ):
            self.type = bool
            self.ranges = {
                k: (v.lower() if isinstance(v, str) else v) in TRUE_BOOLEANS
                for k, v in self.ranges.items()
            }
            self.min_value = False
            self.max_value = True
            self.step_size = 1

        self.valid_values = list(self.ranges.values())
        self.range = self.min_value, self.max_value, self.step_size
        self._value_to_raw_value_ranges = {value: _range for _range, value in self.ranges.items()}

    @contextmanager
    def __get_cpp_parameter(self):
        """
        The C++ version of this class (`_AudioProcessorParameter`) is owned
        by the ExternalPlugin object and is not guaranteed to exist at the
        same memory address every time we might need it. This Python wrapper
        looks it up dynamically.
        """
        _parameter = self.__plugin._get_parameter(self.__parameter_name)  # type: ignore
        if _parameter and _parameter.name == self.__parameter_name:
            yield _parameter
            return
        raise RuntimeError(
            "Parameter {} on plugin {} is no longer available. This could indicate that the plugin"
            " has changed parameters.".format(self.__parameter_name, self.__plugin)
        )

    @property
    def label(self) -> Optional[str]:
        """
        The units used by this parameter (Hz, dB, etc).

        May be ``None`` if the plugin does not expose units for this
        parameter or if automatic unit detection fails.
        """
        if hasattr(self, "_label") and self._label:
            return self._label
        with self.__get_cpp_parameter() as parameter:
            if parameter.label:
                return parameter.label
        return None

    @property
    def units(self) -> Optional[str]:
        """
        Alias for "label" - the units used by this parameter (Hz, dB, etc).

        May be ``None`` if the plugin does not expose units for this
        parameter or if automatic unit detection fails.
        """
        return self.label

    def __repr__(self):
        with self.__get_cpp_parameter() as parameter:
            cpp_repr_value = repr(parameter)
            cpp_repr_value = cpp_repr_value.rstrip(">")
            if self.type is float:
                if self.step_size:
                    return "{} value={} range=({}, {}, {})>".format(
                        cpp_repr_value,
                        self.string_value,
                        self.min_value,
                        self.max_value,
                        self.step_size,
                    )
                elif self.approximate_step_size:
                    return "{} value={} range=({}, {}, ~{})>".format(
                        cpp_repr_value,
                        self.string_value,
                        self.min_value,
                        self.max_value,
                        self.approximate_step_size,
                    )
                else:
                    return "{} value={} range=({}, {}, ?)>".format(
                        cpp_repr_value, self.string_value, self.min_value, self.max_value
                    )
            elif self.type is str:
                return '{} value="{}" ({} valid string value{})>'.format(
                    cpp_repr_value,
                    self.string_value,
                    len(self.valid_values),
                    "" if len(self.valid_values) == 1 else "s",
                )
            elif self.type is bool:
                return '{} value={} boolean ("{}" and "{}")>'.format(
                    cpp_repr_value, self.string_value, self.valid_values[0], self.valid_values[1]
                )
            else:
                raise ValueError(
                    f"Parameter {self.python_name} has an unknown type. (Found '{self.type}')"
                )

    def __getattr__(self, name: str):
        if not name.startswith("_"):
            try:
                with self.__get_cpp_parameter() as parameter:
                    return getattr(parameter, name)
            except RuntimeError:
                pass
        if hasattr(super(), "__getattr__"):
            return super().__getattr__(name)  # type: ignore
        raise AttributeError("'{}' has no attribute '{}'".format(self.__class__.__name__, name))

    def __setattr__(self, name: str, value):
        if not name.startswith("_"):
            try:
                with self.__get_cpp_parameter() as parameter:
                    if hasattr(parameter, name):
                        return setattr(parameter, name, value)
            except RuntimeError:
                pass
        return super().__setattr__(name, value)

    def get_raw_value_for(self, new_value: Union[float, str, bool]) -> float:
        if self.type is float:
            to_float_value: Union[str, float, bool]
            if isinstance(new_value, str) and self.label and new_value.endswith(self.label):
                to_float_value = new_value[: -len(self.label)]
            else:
                to_float_value = new_value

            try:
                new_value = float(to_float_value)
            except ValueError:
                if self.label:
                    raise ValueError(
                        "Value received for parameter '{}' ({}) must be a number or a string (with"
                        " the optional suffix '{}')".format(
                            self.python_name, repr(new_value), self.label
                        )
                    )
                raise ValueError(
                    "Value received for parameter '{}' ({}) must be a number or a string".format(
                        self.python_name, repr(new_value)
                    )
                )
            if (self.min_value is not None and new_value < self.min_value) or (
                self.max_value is not None and new_value > self.max_value
            ):
                raise ValueError(
                    "Value received for parameter '{}' ({}) is out of range [{}{}, {}{}]".format(
                        self.python_name,
                        repr(new_value),
                        self.min_value,
                        self.label,
                        self.max_value,
                        self.label,
                    )
                )
            plugin_reported_raw_value = self.get_raw_value_for_text(str(new_value))

            closest_diff = None
            closest_range_value: Optional[Tuple[float, float]] = None
            for value, raw_value_range in self._value_to_raw_value_ranges.items():
                if not isinstance(value, float):
                    continue
                diff = new_value - value
                if closest_diff is None or abs(diff) < abs(closest_diff):
                    closest_range_value = raw_value_range
                    closest_diff = diff

            if closest_range_value is not None:
                expected_low, expected_high = closest_range_value
                if (
                    plugin_reported_raw_value < expected_low
                    or plugin_reported_raw_value > expected_high
                ):
                    # The plugin might have bad code in it when trying
                    # to parse one of the string values it gave to us.
                    # Let's use the range we had before:
                    return expected_low
            return plugin_reported_raw_value

        elif self.type is str:
            if isinstance(new_value, (str, int, float, bool)):
                new_value = str(new_value)
            else:
                raise ValueError(
                    "Value received for parameter '{}' ({}) should be a string (or string-like),"
                    " but got an object of type: {}".format(
                        self.python_name, repr(new_value), type(new_value)
                    )
                )
            if new_value not in self.valid_values:
                raise ValueError(
                    "Value received for parameter '{}' ({}) not in list of valid values: {}".format(
                        self.python_name, repr(new_value), self.valid_values
                    )
                )
            plugin_reported_raw_value = self.get_raw_value_for_text(new_value)
            expected_low, expected_high = self._value_to_raw_value_ranges[new_value]
            if (
                plugin_reported_raw_value < expected_low
                or plugin_reported_raw_value > expected_high
            ):
                # The plugin might have bad code in it when trying
                # to parse one of the string values it gave to us.
                # Let's use the range we had before:
                return expected_low
            else:
                return plugin_reported_raw_value
        elif self.type is bool:
            if not isinstance(new_value, (bool, WrappedBool)):
                raise ValueError(
                    "Value received for parameter '{}' ({}) should be a boolean,"
                    " but got an object of type: {}".format(
                        self.python_name, repr(new_value), type(new_value)
                    )
                )
            return 1.0 if new_value else 0.0
        else:
            raise ValueError(
                "Parameter has invalid type: {}. This should not be possible!".format(self.type)
            )


def to_python_parameter_name(parameter: _AudioProcessorParameter) -> Optional[str]:
    if not parameter.name and not parameter.label:
        return None

    name = parameter.name
    if parameter.label and not parameter.label.startswith(":"):
        name = "{} {}".format(name, parameter.label)
    return normalize_python_parameter_name(name)


def normalize_python_parameter_name(name: str) -> str:
    name = name.lower().strip()
    # Special case: some plugins expose parameters with "#"/"♯" or "b"/"♭" in their names.
    name = name.replace("#", "_sharp").replace("♯", "_sharp").replace("♭", "_flat")
    # Replace all non-alphanumeric characters with underscores
    name_chars = [
        c if (c.isalpha() or c.isnumeric()) and c.isprintable() and ord(c) < 128 else "_"
        for c in name
    ]
    if not name_chars:  # Unhandled/Weird param name
        return ""
    # Remove any double-underscores:
    name_chars = [a for a, b in zip(name_chars, name_chars[1:]) if a != b or b != "_"] + [
        name_chars[-1]
    ]
    # Remove any leading or trailing underscores:
    name = "".join(name_chars).strip("_")
    return name


class _PythonExternalPluginMixin:
    def __set_initial_parameter_values__(
        self, parameter_values: Optional[Dict[str, Union[str, int, float, bool]]]
    ):
        if parameter_values is None:
            parameter_values = {}
        if not isinstance(parameter_values, dict):
            raise TypeError(
                "Expected a dictionary to be passed to parameter_values, but received a"
                f" {type(parameter_values).__name__}. (If passing a plugin name, pass"
                ' "plugin_name=..." as a keyword argument instead.)'
            )
        parameters = self.parameters
        for key, value in parameter_values.items():
            if key not in parameters:
                raise AttributeError(
                    'Parameter named "{}" not found. Valid options: {}'.format(
                        key,
                        ", ".join(
                            cast(
                                Dict[str, AudioProcessorParameter], self._parameter_weakrefs
                            ).keys()
                        ),
                    )
                )
            setattr(self, key, value)

    @property
    def parameters(self) -> Dict[str, AudioProcessorParameter]:
        # Return a read-only version of this dictionary,
        # to prevent people from trying to assign to it.
        return ReadOnlyDictWrapper(self._get_parameters())

    def _get_parameters(self):
        if not hasattr(self, "__python_parameter_cache__"):
            self.__python_parameter_cache__ = {}
        if not hasattr(self, "__python_to_cpp_names__"):
            self.__python_to_cpp_names__ = {}

        parameters = {}
        for cpp_parameter in self._parameters:  # type: ignore
            if any([regex.match(cpp_parameter.name) for regex in PARAMETER_NAME_REGEXES_TO_IGNORE]):
                continue
            if cpp_parameter.name not in self.__python_parameter_cache__:
                self.__python_parameter_cache__[cpp_parameter.name] = AudioProcessorParameter(
                    self,  # type: ignore
                    cpp_parameter.name,
                )
            parameter = self.__python_parameter_cache__[cpp_parameter.name]
            if parameter.python_name:
                parameters[parameter.python_name] = parameter
                self.__python_to_cpp_names__[parameter.python_name] = cpp_parameter.name
        return parameters

    def _get_parameter_by_python_name(self, python_name: str) -> Optional[AudioProcessorParameter]:
        if not hasattr(self, "__python_parameter_cache__"):
            self.__python_parameter_cache__ = {}
        if not hasattr(self, "__python_to_cpp_names__"):
            self.__python_to_cpp_names__ = {}

        cpp_name = self.__python_to_cpp_names__.get(python_name)
        if not cpp_name:
            return self._get_parameters().get(python_name)

        cpp_parameter = self._get_parameter(cpp_name)  # type: ignore
        if not cpp_parameter:
            return None

        if cpp_parameter.name not in self.__python_parameter_cache__:
            self.__python_parameter_cache__[cpp_parameter.name] = AudioProcessorParameter(
                cast(ExternalPlugin, self), cpp_parameter.name
            )
        return self.__python_parameter_cache__[cpp_parameter.name]

    def __dir__(self):
        parameter_names = []
        for parameter in self._parameters:  # type: ignore
            name = to_python_parameter_name(parameter)
            if name:
                parameter_names.append(name)
        return super().__dir__() + parameter_names  # type: ignore

    def __getattr__(self, name: str):
        if not name.startswith("_"):
            parameter = self._get_parameter_by_python_name(name)
            if parameter:
                string_value = parameter.string_value
                if parameter.type is float:
                    return FloatWithParameter(
                        float(strip_common_float_suffixes(string_value)),
                        wrapped=parameter,  # type: ignore
                    )  # type: ignore
                elif parameter.type is bool:
                    return BooleanWithParameter(
                        parameter.raw_value >= 0.5,
                        wrapped=parameter,  # type: ignore
                    )  # type: ignore
                elif parameter.type is str:
                    return StringWithParameter(str(string_value), wrapped=parameter)  # type: ignore
                else:
                    raise ValueError(
                        f"Parameter {parameter.python_name} has an unknown type. (Found"
                        f" '{parameter.type}')"
                    )
        return getattr(super(), name)

    def __setattr__(self, name: str, value):
        if not name.startswith("__"):
            parameter = self._get_parameter_by_python_name(name)
            if parameter:
                parameter.raw_value = parameter.get_raw_value_for(value)
                return
        super().__setattr__(name, value)


ExternalPlugin.__bases__ = ExternalPlugin.__bases__ + (_PythonExternalPluginMixin,)


_AVAILABLE_PLUGIN_CLASSES: List[Type[ExternalPlugin]] = []

try:
    from pedalboard_native import VST3Plugin  # type: ignore

    _AVAILABLE_PLUGIN_CLASSES.append(VST3Plugin)
except ImportError:
    # We may be on a system that doesn't have native VST3Plugin support.
    pass

try:
    from pedalboard_native import AudioUnitPlugin  # type: ignore

    _AVAILABLE_PLUGIN_CLASSES.append(AudioUnitPlugin)
except ImportError:
    # We may be on a system that doesn't have native AudioUnitPlugin support.
    # (i.e.: any platform that's not macOS.)
    pass


def load_plugin(
    path_to_plugin_file: str,
    parameter_values: Dict[str, Union[str, int, float, bool]] = {},
    plugin_name: Union[str, None] = None,
    initialization_timeout: float = 10.0,
) -> ExternalPlugin:
    """
    Load an audio plugin.

    Two plugin formats are supported:
     - VST3® format is supported on macOS, Windows, and Linux
     - Audio Units are supported on macOS

    Args:
        path_to_plugin_file (``str``): The path of a VST3® or Audio Unit plugin file or bundle.

        parameter_values (``Dict[str, Union[str, int, float, bool]]``):
            An optional dictionary of initial values to provide to the plugin
            after loading. Keys in this dictionary are expected to match the
            parameter names reported by the plugin, but normalized to strings
            that can be used as Python identifiers. (These are the same
            identifiers that are used as keys in the ``.parameters`` dictionary
            of a loaded plugin.)

        plugin_name (``Optional[str]``):
            An optional plugin name that can be used to load a specific plugin
            from a multi-plugin package. If a package is loaded but a
            ``plugin_name`` is not provided, an exception will be thrown.

        initialization_timeout (``float``):
            The number of seconds that Pedalboard will spend trying to load this plugin.
            Some plugins load resources asynchronously in the background on startup;
            using larger values for this parameter can give these plugins time to
            load properly.

            *Introduced in v0.7.6.*

    Returns:
        an instance of :class:`pedalboard.VST3Plugin` or :class:`pedalboard.AudioUnitPlugin`

    Throws:
        ``ImportError``: if the plugin cannot be found or loaded

        ``RuntimeError``: if the plugin file contains more than one plugin,
        but no ``plugin_name`` was provided
    """
    if not _AVAILABLE_PLUGIN_CLASSES:
        raise ImportError(
            "Pedalboard found no supported external plugin types in this installation ({}).".format(
                platform.system()
            )
        )
    exceptions = []
    for plugin_class in _AVAILABLE_PLUGIN_CLASSES:
        try:
            return plugin_class(  # type: ignore
                path_to_plugin_file=path_to_plugin_file,  # type: ignore
                parameter_values=parameter_values,  # type: ignore
                plugin_name=plugin_name,  # type: ignore
                initialization_timeout=initialization_timeout,  # type: ignore
            )  # type: ignore
        except ImportError as e:
            exceptions.append(e)
        except Exception:
            raise
    else:
        tried_plugins = ", ".join([c.__name__ for c in _AVAILABLE_PLUGIN_CLASSES])
        # Good error messages are important, okay?
        if len(_AVAILABLE_PLUGIN_CLASSES) > 2:
            tried_plugins = ", or ".join(tried_plugins.rsplit(", ", 1))
        else:
            tried_plugins = " or ".join(tried_plugins.rsplit(", ", 1))
        raise ImportError(
            "Failed to load plugin as {}. Errors were:\n\t{}".format(
                tried_plugins,
                "\n\t".join(
                    [
                        "{}: {}".format(klass.__name__, exception)
                        for klass, exception in zip(_AVAILABLE_PLUGIN_CLASSES, exceptions)
                    ]
                ),
            )
        )
